@@ -25,6 +25,7 @@
 #include <glib.h>
 #include <gmodule.h>
 #include <string.h>
+#include <stdlib.h>
 
 #include <libmafw/mafw.h>
 #include <grilo.h>
@@ -36,6 +37,8 @@
 
 #define MAFW_GRILO_SOURCE_PLUGIN_NAME "MAFW-Grilo-Source"
 
+#define MAX_COUNT 1024
+
 
 G_DEFINE_TYPE (MafwGriloSource, mafw_grilo_source, MAFW_TYPE_SOURCE);
 
@@ -84,7 +87,11 @@ typedef struct
   gpointer mafw_user_data;
   guint mafw_browse_id;
   guint grl_browse_id;
-  guint index;
+  guint item_count;
+  gint total_items;
+  gboolean more_pages;
+  GrlMedia *grl_media;
+  guint pagination_skip;
 } BrowseCbInfo;
 
 typedef struct
@@ -262,6 +269,10 @@ destroy_browse_cb_info (gpointer user_data)
   BrowseCbInfo *browse_cb_info = user_data;
 
   g_object_unref (browse_cb_info->mafw_grilo_source);
+  if (browse_cb_info->grl_media)
+    {
+      g_object_unref (browse_cb_info->grl_media);
+    }
   g_free (browse_cb_info);
 }
 
@@ -570,41 +581,67 @@ mafw_grilo_source_new (GrlMediaPlugin *grl_plugin)
                        NULL);
 }
 
-static GrlMedia *
-grl_media_deserialize (const gchar *object_id)
+static void
+grl_media_deserialize (const gchar *object_id, GrlMedia **grl_media,
+                       guint *pagination_skip)
 {
-  GrlMedia *grl_media = NULL;
   gchar *serialized_grl_media = NULL;
-  gchar *grl_media_type, *grl_media_id;
 
   if (mafw_source_split_objectid (object_id, NULL, &serialized_grl_media) &&
       serialized_grl_media[0] != '\0')
     {
+      guint skip;
+      gchar *grl_media_type;
+      gchar *ptr;
+
+      skip = strtoul (serialized_grl_media, &ptr, 10);
+
+      if (pagination_skip)
+        {
+          *pagination_skip = skip;
+        }
+
       /* We search ':' and then we convert the type in NULL terminated and
          prepare the media id */
-      grl_media_type = serialized_grl_media;
-      grl_media_id = g_strstr_len (serialized_grl_media, -1, ":");
-      grl_media_type[grl_media_id - grl_media_type] = '\0';
-      grl_media_id++;
+      grl_media_type = ptr + 1;
+      if (grl_media_type[0] != '\0')
+        {
+          gchar *grl_media_id;
 
-      grl_media = g_object_new (g_type_from_name (grl_media_type), NULL);
-      grl_media_set_id (grl_media, grl_media_id);
+          grl_media_id = g_strstr_len (grl_media_type, -1, ":");
+          grl_media_type[grl_media_id - grl_media_type] = '\0';
+          grl_media_id++;
+
+          *grl_media =
+            g_object_new (g_type_from_name (grl_media_type), NULL);
+          grl_media_set_id (*grl_media, grl_media_id);
+        }
     }
 
   g_free (serialized_grl_media);
-
-  return grl_media;
 }
 
 static gchar *
-grl_media_serialize (GrlMedia *grl_media, const gchar *source_id)
+grl_media_serialize (GrlMedia *grl_media, const gchar *source_id,
+                     guint pagination_skip)
 {
   const gchar *media_id, *type;
+  gchar *value;
 
-  type = G_OBJECT_TYPE_NAME (grl_media);
-  media_id = grl_media_get_id (grl_media);
+  if (grl_media)
+    {
+      type = G_OBJECT_TYPE_NAME (grl_media);
+      media_id = grl_media_get_id (grl_media);
 
-  return g_strconcat (source_id, "::", type, ":", media_id, NULL);
+      value = g_strdup_printf ("%s::%u:%s:%s", source_id, pagination_skip, type,
+                               media_id);
+    }
+  else
+    {
+      value = g_strdup_printf ("%s::%u:", source_id, pagination_skip);
+    }
+
+  return value;
 }
 
 static GList *
@@ -758,6 +795,55 @@ mafw_keys_from_grl_media (MafwGriloSource *mafw_source, GrlMedia *grl_media)
   return mafw_metadata_keys;
 }
 
+static GHashTable *
+get_next_row_metadata_keys (void)
+{
+  GHashTable *keys;
+
+  keys = mafw_metadata_new ();
+
+  mafw_metadata_add_str (keys, MAFW_METADATA_KEY_TITLE, "More results...");
+  mafw_metadata_add_str (keys, MAFW_METADATA_KEY_MIME,
+                         MAFW_METADATA_VALUE_MIME_CONTAINER);
+
+  return keys;
+}
+
+static gboolean
+add_next_page_row (gpointer user_data)
+{
+  BrowseCbInfo *browse_cb_info = user_data;
+  gchar *object_id;
+  GHashTable *mafw_metadata_keys;
+
+  object_id =
+    grl_media_serialize (browse_cb_info->grl_media,
+                         mafw_extension_get_uuid (MAFW_EXTENSION (browse_cb_info->mafw_grilo_source)),
+                         browse_cb_info->pagination_skip);
+
+  mafw_metadata_keys = get_next_row_metadata_keys ();
+
+  browse_cb_info->mafw_browse_cb (MAFW_SOURCE (browse_cb_info->
+                                               mafw_grilo_source),
+                                  browse_cb_info->mafw_browse_id, 0, 0,
+                                  object_id, mafw_metadata_keys,
+                                  browse_cb_info->mafw_user_data, NULL);
+
+  g_free (object_id);
+  if (mafw_metadata_keys)
+    {
+      g_hash_table_unref (mafw_metadata_keys);
+    }
+
+  /* we don't free the info, we just remove it from the hash table
+     and it will free it for us */
+  g_hash_table_remove (browse_cb_info->mafw_grilo_source->priv->
+                       browse_requests,
+                       &(browse_cb_info->mafw_browse_id));
+
+  return FALSE;
+}
+
 static void
 grl_browse_cb (GrlMediaSource *grl_source,
                guint grl_browse_id,
@@ -770,6 +856,8 @@ grl_browse_cb (GrlMediaSource *grl_source,
   gchar *mafw_object_id = NULL;
   GHashTable *mafw_metadata_keys = NULL;
 
+  browse_cb_info->more_pages |= remaining + 1 >= MAX_COUNT;
+
   if (grl_media)
     {
       const gchar *mafw_uuid;
@@ -778,17 +866,18 @@ grl_browse_cb (GrlMediaSource *grl_source,
                                                            mafw_grilo_source));
 
       mafw_object_id =
-        grl_media_serialize (grl_media, mafw_uuid);
+        grl_media_serialize (grl_media, mafw_uuid, 0);
       mafw_metadata_keys = mafw_keys_from_grl_media (browse_cb_info->
                                                      mafw_grilo_source,
                                                      grl_media);
+      browse_cb_info->total_items++;
     }
 
   browse_cb_info->mafw_browse_cb (MAFW_SOURCE (browse_cb_info->
                                                mafw_grilo_source),
                                   browse_cb_info->mafw_browse_id,
-                                  remaining,
-                                  browse_cb_info->index,
+                                  browse_cb_info->more_pages ?
+                                  remaining + 1 : remaining, 0,
                                   mafw_object_id,
                                   mafw_metadata_keys,
                                   browse_cb_info->mafw_user_data,
@@ -802,11 +891,26 @@ grl_browse_cb (GrlMediaSource *grl_source,
 
   if (!remaining || error)
     {
-      /* we don't free the info, we just remove it from the hash table
-         and it will free it for us */
-      g_hash_table_remove (browse_cb_info->mafw_grilo_source->priv->
-                           browse_requests,
-                           &(browse_cb_info->mafw_browse_id));
+      if (browse_cb_info->total_items > browse_cb_info->item_count)
+        {
+          g_message ("Exceeded item_count %d by returning %d items",
+                     browse_cb_info->item_count,
+                     browse_cb_info->total_items);
+        }
+
+      if (browse_cb_info->more_pages)
+        {
+          browse_cb_info->pagination_skip += MAX_COUNT;
+          g_idle_add (add_next_page_row, browse_cb_info);
+        }
+      else
+        {
+          /* we don't free the info, we just remove it from the hash table
+             and it will free it for us */
+          g_hash_table_remove (browse_cb_info->mafw_grilo_source->priv->
+                               browse_requests,
+                               &(browse_cb_info->mafw_browse_id));
+        }
     }
 }
 
@@ -876,7 +980,7 @@ mafw_grilo_source_browse (MafwSource *source,
                           MafwSourceBrowseResultCb browse_cb,
                           gpointer user_data)
 {
-  GrlMedia *grl_media;
+  GrlMedia *grl_media = NULL;
   BrowseCbInfo *browse_cb_info;
   GList *grl_keys;
 
@@ -889,8 +993,12 @@ mafw_grilo_source_browse (MafwSource *source,
   browse_cb_info->mafw_user_data = user_data;
   browse_cb_info->mafw_browse_id =
     browse_cb_info->mafw_grilo_source->priv->next_browse_id++;
+  browse_cb_info->item_count = item_count;
 
-  grl_media = grl_media_deserialize (object_id);
+  grl_media_deserialize (object_id, &grl_media,
+                         &(browse_cb_info->pagination_skip));
+
+  browse_cb_info->grl_media = grl_media ? g_object_ref (grl_media) : NULL;
 
   grl_keys = mafw_keys_to_grl_keys (MAFW_GRILO_SOURCE (source), metadata_keys);
 
@@ -904,8 +1012,8 @@ mafw_grilo_source_browse (MafwSource *source,
                                                grl_source),
                              grl_media,
                              grl_keys,
-                             skip_count,
-                             item_count ? item_count : G_MAXUINT,
+                             browse_cb_info->pagination_skip,
+                             MAX_COUNT,
                              GRL_RESOLVE_IDLE_RELAY |
                              browse_cb_info->mafw_grilo_source->priv->
                              browse_metadata_mode,
@@ -958,7 +1066,7 @@ mafw_grilo_source_get_metadata (MafwSource *source,
                                 gpointer user_data)
 {
   MetadataCbInfo *metadata_cb_info;
-  GrlMedia *grl_media;
+  GrlMedia *grl_media = NULL;
   GList *grl_keys;
   GrlSupportedOps supported_ops;
 
@@ -971,7 +1079,7 @@ mafw_grilo_source_get_metadata (MafwSource *source,
   metadata_cb_info->mafw_user_data = user_data;
   metadata_cb_info->mafw_object_id = g_strdup (object_id);
 
-  grl_media = grl_media_deserialize (object_id);
+  grl_media_deserialize (object_id, &grl_media, NULL);
   grl_keys = mafw_keys_to_grl_keys (MAFW_GRILO_SOURCE (source), metadata_keys);
 
   supported_ops =
